###################################################################################
#
#   LemonLDAP::Handlers::NTLM - Copyright (c) 2007 Sebastien DIAZ
#
#   You may distribute under the terms of either the GNU General Public
#   License or the Artistic License, as specified in the Perl README file.
#
#   THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
#   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#   WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
#
###################################################################################


package Lemonldap::Handlers::NTLM ;

#use strict ;
use vars qw{$cache $VERSION %msgflags1 %msgflags2 %msgflags3 %invflags1 %invflags2 %invflags3 $addr $port $debug} ;

$VERSION = 0.02 ;

$debug = 0 ;

$cache = undef ;
use Apache2::RequestUtil;
use MIME::Base64 () ;
use Authen::Smb 0.95 ;
use Socket ;
use Apache2::Log();
use Apache2::ServerRec();

%msgflags1 = ( 0x01 => "NEGOTIATE_UNICODE",
               0x02 => "NEGOTIATE_OEM",
               0x04 => "REQUEST_TARGET",
               0x10 => "NEGOTIATE_SIGN",
               0x20 => "NEGOTIATE_SEAL",
               0x80 => "NEGOTAITE_LM_KEY",
               );

%msgflags2 = ( 0x02 => "NEGOTIATE_NTLM",
               0x40 => "NEGOTIATE_LOCAL_CALL",
               0x80 => "NEGOTIATE_ALWAYS_SIGN",
               );

%msgflags3 = ( 0x01 => "TARGET_TYPE_DOMAIN",
               0x02 => "TARGET_TYPE_SERVER",
               );

%invflags1 = ( "NEGOTIATE_UNICODE" => 0x01,
               "NEGOTIATE_OEM"     => 0x02,
               "REQUEST_TARGET"    => 0x04,
               "NEGOTIATE_SIGN"    => 0x10,
               "NEGOTIATE_SEAL"    => 0x20,
               "NEGOTIATE_LM_KEY"  => 0x80,
               );

%invflags2 = ( "NEGOTIATE_NTLM"        => 0x02,
               "NEGOTIATE_LOCAL_CALL"  => 0x40,
               "NEGOTIATE_ALWAYS_SIGN" => 0x80,
               );

%invflags3 = ( "TARGET_TYPE_DOMAIN" => 0x01,
               "TARGET_TYPE_SERVER" => 0x02,
               );

############################################
# here is where we start the new code....
############################################
use mod_perl2 ;

use Apache2::Access ;
use Apache2::Connection ;
use Apache2::Log ;
use Apache2::RequestRec ;
use Apache2::RequestUtil ;
use Apache2::RequestIO ;
use APR::Table ;
use APR::SockAddr ;
use Apache2::Const -compile => qw(HTTP_UNAUTHORIZED HTTP_INTERNAL_SERVER_ERROR DECLINED HTTP_FORBIDDEN OK) ;


###########################################
#  Global Parameters
##########################################
my @ListNoProfilServer;
my $debug;
##################### end modperl code ######################

sub get_config
{
    my ($self, $r) = @_ ;

    return if ($self -> {smbpdc}) ; # config already setup

    $debug = $r -> dir_config ('ntlmdebug') || 0 ;
    $debug = $self -> {debug} = lc($debug) eq 'on'?1:($debug+0) ;

    my @config = $r -> dir_config -> get ('ntdomain') ;

    foreach (@config)
    {
        my ($domain, $pdc, $bdc) = split /\s+/ ;
        $domain = lc ($domain) ;
        $self -> {smbpdc}{$domain} = $pdc ;
        $self -> {smbbdc}{$domain} = $bdc ;
        #print STDERR "[$$] AuthenNTLM: Config Domain = $domain  pdc = $pdc  bdc = $bdc\n" if ($debug) ;
    }

    $self -> {defaultdomain} = $r -> dir_config ('defaultdomain') || '' ;
    $self -> {fallbackdomain} = $r -> dir_config ('fallbackdomain') || '' ;
    $self -> {cacheuser} = $r -> dir_config ('cacheuser') || '0' ;
    $self -> {authtype} = $r -> auth_type || 'ntlm,basic' ;
    $self -> {authname} = $r -> auth_name || ''  ;
    my $autho = $r -> dir_config ('ntlmauthoritative') || 'on' ;
    $self -> {ntlmauthoritative} = lc($autho) eq 'on' || $autho == 1?1:0 ;
    $autho = $r -> dir_config ('basicauthoritative') || 'on' ;
    $self -> {basicauthoritative} = lc($autho) eq 'on' || $autho == 1?1:0 ;

    $self -> {authntlm} = 0 ;
    $self -> {authbasic} = 0 ;

    $self -> {authntlm} = 1 if ($self -> {authtype} =~ /(^|,)ntlm($|,)/i) ;
    $self -> {authbasic} = 1 if ($self -> {authtype} =~ /(^|,)basic($|,)/i) ;

    $self -> {semkey} = $r -> dir_config ('ntlmsemkey') ;
    $self -> {semkey} = 23754 if (!defined ($self -> {semkey})) ;
    $self -> {semtimeout} = $r -> dir_config ('ntlmsemtimeout') ;
    $self -> {semtimeout} = 2 if (!defined ($self -> {semtimeout})) ;
    $self -> {splitdomainprefix} = $r -> dir_config ('splitdomainprefix') || '' ;

    if ($debug)
    {
        print STDERR "[$$] AuthenNTLM: Config Default Domain = $self->{defaultdomain}\n"  ;
        print STDERR "[$$] AuthenNTLM: Config Fallback Domain = $self->{fallbackdomain}\n"  ;
        print STDERR "[$$] AuthenNTLM: Config AuthType = $self->{authtype} AuthName = $self->{authname}\n"  ;
        print STDERR "[$$] AuthenNTLM: Config Auth NTLM = $self->{authntlm} Auth Basic = $self->{authbasic}\n"  ;
        print STDERR "[$$] AuthenNTLM: Config NTLMAuthoritative = " .  ($self -> {ntlmauthoritative}?'on':'off') . "  BasicAuthoritative = " . ($self -> {basicauthoritative}?'on':'off') . "\n"  ;
        print STDERR "[$$] AuthenNTLM: Config Semaphore key = $self->{semkey} timeout = $self->{semtimeout}\n"  ;
        print STDERR "[$$] AuthenNTLM: Config SplitDomainPrefix = $self->{splitdomainprefix}\n"  ;
    }
}


sub get_nonce
{
    my ($self, $r) = @_ ;

    # reuse connection if possible
    if ($self -> {nonce} && $self -> {ok})
    {
        #print STDERR "[$$] AuthenNTLM: Returning cached nonce\n" if ($debug) ;
        return $self -> {nonce}
    }

    # this is not the real nonce!
    # we just need to preallocate some space (8 bytes) where Authen::Smb::Valid_User_Connect
    # puts the real nonce
    my $nonce = '12345678' ;
    #print STDERR "Domain".$self -> {domain}."\n";

    $self -> {domain} = $self -> {fallbackdomain}
    if ($self -> {fallbackdomain} && !($self -> {smbpdc}{lc ($self -> {domain})}));
    #print STDERR "Domain".$self -> {domain}."\n";
    my $domain  = lc ($self -> {domain}) ;
    my $pdc     = $self -> {smbpdc}{$domain} ;
    my $bdc     = $self -> {smbbdc}{$domain} ;
    $domain  = lc ($self -> {domain}) ;
    $pdc     = $self -> {smbpdc}{$domain} ;
    $bdc     = $self -> {smbbdc}{$domain} ;
    if (!$pdc)
    {
        $r->log_error("No PDC and no fallbackdomain given for domain $domain") ;
        return '' ;
    }

    $self -> {nonce} = undef ;

    #print STDERR "[$$] AuthenNTLM: Connect to pdc = $pdc bdc = $bdc domain = $domain\n" if ($debug) ;

    # smb aborts any connection that where no user is logged on as soon as somebody
    # tries to open another one. So we have to make sure two request, do not start
    # two auth cycles at the same time. To avoid a hang of the whole server we wrap it with
    # a small timeout
    if ($self->{semkey})
    {
        eval
        {
            local $SIG{ALRM} = sub { #print STDERR "[$$] AuthenNTLM: timed out"
                                #        . "while waiting for lock (key = $self->{semkey})\n";
                                die ; };

            alarm $self -> {semtimeout} ;
            $self -> {lock} = Lemonldap::Handlers::LockNTLM -> lock ($self->{semkey}, $debug) ;
            alarm 0;
        };
    }

    $self -> {smbhandle} = Authen::Smb::Valid_User_Connect ($pdc, $bdc, $domain, $nonce) ;

    #print STDERR "[$$] AuthenNTLM: verify handle $self->{username} smbhandle == $self->{smbhandle} \n" if ($debug) ;

    if (!$self -> {smbhandle})
    {
        $r->log_error("Connect to SMB Server failed (pdc = $pdc bdc = $bdc domain = $domain error = "
                      . Authen::Smb::SMBlib_errno . '/' . Authen::Smb::SMBlib_SMB_Error . ") for " .
                      $r -> uri) ;
        return undef ;
    }

    return $self -> {nonce} = $nonce ;
}



sub verify_user
{
    my ($self, $r) = @_ ;

    if (!$self -> {smbhandle})
    {
        $self -> {lock} = undef ; # reset lock in case anything has gone wrong
        $r->log_error("SMB Server connection not open in state 3 for " . $r -> uri) ;
        return ;
    }

    my $rc ;

#   print STDERR "[$$] AuthenNTLM: Verify user $self->{username} via smb server\n" if ($debug) ;

    if ($self -> {basic})
    {
        $rc = Authen::Smb::Valid_User_Auth ($self -> {smbhandle}, $self->{username}, $self -> {password}) ;
    }
    else
    {
        $rc = Authen::Smb::Valid_User_Auth ($self -> {smbhandle}, $self->{username}, $self -> {usernthash}, 1, $self->{userdomain}) ;
    }
    my $errno  = Authen::Smb::SMBlib_errno ;
    my $smberr = Authen::Smb::SMBlib_SMB_Error ;
    Authen::Smb::Valid_User_Disconnect ($self -> {smbhandle}) if ($self -> {smbhandle}) ;
    $self -> {smbhandle} = undef ;
    $self -> {lock}      = undef ;

    if ($rc == &Authen::Smb::NTV_LOGON_ERROR)
    {
        $r->log_error("Wrong password/user (rc=$rc/$errno/$smberr): $self->{userdomain}\\$self->{username} for " . $r -> uri) ;
#       print STDERR "[$$] AuthenNTLM: rc = $rc  ntlmhash = $self->{usernthash}\n" if ($debug) ;
        return ;
    }

    if ($rc)
    {
        $r->log_reason("SMB Server error $rc/$errno/$smberr for " . $r -> uri) ;
        return ;
    }

    return 1 ;
}


sub map_user
{
    my ($self, $r) = @_ ;

    if ($self -> {splitdomainprefix} == 1)
    {
        return lc("$self->{username}") ;
    }
    else
    {
        return lc("$self->{userdomain}\\$self->{username}") ;
    }
}



sub substr_unicode
{
    my ($data, $off,  $len) = @_ ;

    my $i = 0 ;
    my $end = $off + $len ;
    my $result = '' ;
    for ($i = $off ; $i < $end ; $i += 2)
    {# for now we simply ignore high order byte
         $result .=  substr ($data, $i,  1) ;
    }

    return $result ;
}


sub get_msg_data
{
    my ($self, $r) = @_ ;

    my $auth_line =  $r->headers_in->{$r->proxyreq ? 'Proxy-Authorization' : 'Authorization'} ;

    $self -> {ntlm}  = 0 ;
    $self -> {basic} = 0 ;
    if ($debug)
    {
        $auth_line =~ /^(.*?)\s+/ ;
        my $type = $1 ;
  #      print STDERR "[$$] AuthenNTLM: Authorization Header "
#           . (defined($auth_line)?($debug > 1?$auth_line:$type):'<not given>') . "\n" if ($debug) ;
    }
    if ($self -> {authntlm} && ($auth_line =~ /^NTLM\s+(.*?)$/i))
    {
        $self -> {ntlm} = 1 ;
    }
    elsif ($self -> {authbasic} && ($auth_line =~ /^Basic\s+(.*?)$/i))
    {
        $self -> {basic}  = 1 ;
    }
    else
    {
        return undef ;
    }

    my $data = MIME::Base64::decode($1) ;

    if ($debug > 1)
    {
        my @out ;
        for (my $i = 0; $i < length($data); $i++)
        {
            push @out, unpack('C', substr($data, $i, 1)) ;
        }
#       print STDERR "[$$] AuthenNTLM: Got: " . join (' ', @out). "\n" ;
    }

    return $data ;
}



sub get_msg
{
    my ($self, $r) = @_ ;

    my $data = $self -> get_msg_data ($r) ;
    return undef if (!$data) ;

    if ($self -> {ntlm})
    {
        my ($protocol, $type) = unpack ('Z8C', $data) ;
        return $self -> get_msg1 ($r, $data) if ($type == 1) ;
        return $self -> get_msg3 ($r, $data) if ($type == 3) ;
        return $type ;
    }
    elsif ($self -> {basic})
    {
        return $self -> get_basic ($r, $data) ;
    }
    return undef ;
}



sub get_msg1
{
    my ($self, $r, $data) = @_ ;

    my ($protocol, $type, $zero, $flags1, $flags2, $zero2, $dom_len, $x1, $dom_off, $x2, $host_len, $x3, $host_off, $x4) = unpack ('Z8Ca3CCa2vvvvvvvv', $data) ;
    my $host   = $host_off?substr ($data, $host_off, $host_len):'' ;
    my $domain = $dom_off?substr ($data, $dom_off,  $dom_len):'' ;

    $self -> {domain} = $dom_len?$domain:$self -> {defaultdomain} ;
    $self -> {host}   = $host_len?$host:'' ;

    $self -> {accept_unicode} = $flags1 & 0x01;

    if ($debug)
    {
        my @flag1str;
        foreach my $i ( sort keys %msgflags1 )
        {
            push @flag1str, $msgflags1{ $i } if $flags1 & $i;
        }
        my $flag1str = join( ",", @flag1str );

        my @flag2str;
        foreach my $i ( sort keys %msgflags2 )
        {
            push @flag2str, $msgflags2{ $i } if $flags2 & $i;
        }
        my $flag2str = join( ",", @flag2str );

#       print STDERR "[$$] AuthenNTLM: protocol=$protocol, type=$type, flags1=$flags1($flag1str), "
#           . "flags2=$flags2($flag2str), domain length=$dom_len, domain offset=$dom_off, "
#           . "host length=$host_len, host offset=$host_off, host=$host, domain=$domain\n" ;
    }


    return $type ;
}

sub hex_to_ascii ($)
    {
        ## Convert each two-digit hex number back to an ASCII character.
        (my $str = shift) =~ s/([a-fA-F0-9]{2})/chr(hex $1)/eg;
        return $str;
    }

sub ascii_to_hex ($)
    {
        ## Convert each ASCII character to a two-digit hex number.
        (my $str = shift) =~ s/(.|\n)/sprintf("%02lx", ord $1)/eg;
        return $str;
    }


sub set_msg2
{
    my ($self, $r, $nonce) = @_ ;

    my $charencoding = $self->{ accept_unicode } ? $invflags1{ NEGOTIATE_UNICODE } : $invflags1{ NEGOTIATE_OEM };

    my $flags2 = $invflags2{ NEGOTIATE_ALWAYS_SIGN } | $invflags2{ NEGOTIATE_NTLM };

    my $data = pack ('Z8Ca7vvCCa2a8a8', 'NTLMSSP', 2, '', 40, 0, $charencoding,  $flags2, '', $nonce, '') ;
    #my $data = 'NTLMSSP '.0x02.0x00.0x0.0x00.'123456789012'.hex_to_ascii (0x00000000).'12345678' ;

#    print STDERR "DataBetween:".$charencoding."\n";
#    print STDERR "DataBetween:".$flags2."\n";
#    print STDERR "DataBetween:".$nonce."\n";
#    print STDERR "DataBetween:".$data."\n";
    my $header = 'NTLM '. MIME::Base64::encode($data, '') ;

    $header = 'NTLM '. MIME::Base64::encode_base64($data, '') ;
    if ($debug)
    {
        if ($debug > 1)
        {
            my @out ;
            for (my $i = 0; $i < length($data); $i++)
            {
                push @out, unpack('C', substr($data, $i, 1)) ;
            }
#           print STDERR "[$$] AuthenNTLM: Send: " . join (' ', @out). "\n" ;
        }
  #      print STDERR "[$$] AuthenNTLM: charencoding = $charencoding\n";
   #     print STDERR "[$$] AuthenNTLM: flags2 = $flags2\n";
  #      print STDERR "[$$] AuthenNTLM: nonce=$nonce\n" if ($debug > 1);
  #      print STDERR "[$$] AuthenNTLM: Send header: " . ($debug > 1?$header:'NTLM ...') . "\n" ;
    }

    return $header;
}


sub get_msg3
{
    my ($self, $r, $data) = @_ ;

    my ($protocol, $type, $zero,
        $lm_len,  $l1, $lm_off,
        $nt_len,   $l3, $nt_off,
        $dom_len, $x1, $dom_off,
        $user_len, $x3, $user_off,
        $host_len, $x5, $host_off,
        $msg_len
        ) = unpack ('Z8Ca3vvVvvVvvVvvVvvVv', $data) ;

    my $lm     = $lm_off  ? substr ($data, $lm_off,   $lm_len):'' ;
    my $nt     = $nt_off  ? substr ($data, $nt_off,   $nt_len):'' ;
    my $user   = $user_off ? ($self->{accept_unicode} ? substr_unicode ($data, $user_off, $user_len) : substr( $data, $user_off, $user_len ) ) :'' ;
    my $host   = $host_off ? ($self->{accept_unicode} ? substr_unicode ($data, $host_off, $host_len) : substr( $data, $host_off, $host_len ) ) :'' ;
    my $domain = $dom_off ? ($self->{accept_unicode} ? substr_unicode ($data, $dom_off,  $dom_len) : substr( $data, $dom_off, $dom_len ) ) :'' ;

    $self -> {userdomain} = $dom_len?$domain:$self -> {defaultdomain} ;
    $self -> {username}   = $user ;
    $self -> {usernthash} = $nt_len ? $nt : $lm;

    if ($debug)
    {
        print STDERR "[$$] AuthenNTLM: protocol=$protocol, type=$type, user=$user, "
            . "host=$host, domain=$domain, msg_len=$msg_len\n" ;
    }


    return $type ;
}


sub get_basic
{
    my ($self, $r, $data) = @_ ;

    ($self -> {username}, $self -> {password}) = split (/:/, $data)  ;

    my ($domain, $username) = split (/\\|\//, $self -> {username}) ;
    if ($username)
    {
        $self -> {domain} = $domain ;
        $self -> {username} = $username ;
    }
    else
    {
        $self -> {domain} = $self -> {defaultdomain} ;
    }

    $self -> {userdomain} = $self -> {domain} ;

    if ($debug)
    {
        print STDERR "[$$] AuthenNTLM: basic auth username = $self->{domain}\\$self->{username}\n" ;
    }

    return -1 ;
}


sub DESTROY
{
    my ($self) = @_ ;

    Authen::Smb::Valid_User_Disconnect ($self -> {smbhandle}) if ($self -> {smbhandle}) ;
}


#sub handler : method
sub NTLMAuth
{

    my ($class, $r) = @_ ;
    $r = Apache2::RequestUtil->request;

   ###############################
   # If Cookie LemonLDAP present no handler
   ##############################
   my $TheCookies="";
   $TheCookies=$r->header_in("Cookie");

    if ( $TheCookies=~/'.*lemonldap.*'/)
    {
        return Apache2::Const::OK ;
    }
    my $type ;
    my $nonce = '' ;
    my $self ;
    my $conn = $r -> connection ;
    my $connhdr = $r -> headers_in -> {'Connection'} ;
    ####################################################
    # Global Properties
    ###################################################
    @ListNoProfilServer = $r->dir_config('ListNoProfilServer');
     my $userisOK="false";



    my $fh = select (STDERR) ;
    $| = 1 ;
    select ($fh) ;

    my $addr = $conn -> remote_addr -> ip_get ;
    my $port = $conn -> remote_addr -> port ;

#   print STDERR "[$$] AuthenNTLM: Start NTLM Authen handler pid = $$, connection = "
#       . "$$conn conn_http_hdr = $connhdr  main = " . ($r -> main)
#       . " cuser = " . $r -> user . ' remote_ip = ' . $conn -> remote_ip
#       . " remote_port = " . unpack('n', $port) . ' remote_host = <'
#       . $conn -> remote_host . "> version = $VERSION "
#       . "smbhandle = " . $self -> {smbhandle} . "\n" if ($debug) ;

    # we cannot attach our object to the connection record. Since in
    # Apache 1.3 there is only one connection at a time per process
    # we can cache our object and check if the connection has changed.
    # The check is done by slightly changing the remote_host member, which
    # persists as long as the connection does
    # This has to be reworked to work with Apache 2.0
    my $table;
    $table = $conn->notes();
    if (1 || ref ($cache) ne $class || $$conn != $cache->{connectionid} || ($table->get('status') ne "AUTHSTARTED"))
    {
        $table->add('status','AUTHSTARTED');
        $conn->notes($table);
        $self = {connectionid => $$conn } ;
        bless $self, $class ;
        $cache = $self ;
#       print STDERR "[$$] AuthenNTLM: Setup new object\n" if ($debug) ;
    }
    else
    {
        $self = $cache ;
#       print STDERR "[$$] AuthenNTLM: Object exists user = $self->{userdomain}\\$self->{username}\n" if ($debug) ;

        if ($self -> {ok})
        {
            $r -> user($self->{mappedusername}) ;

            # we accept the user because we are on the same connection
            $type = $self -> get_msg ($r);
            my $content_len = $r->headers_in->{'content-length'} ;
            my $method      = $r -> method ;
            #print STDERR "[$$] AuthenNTLM: Same connection pid = $$, connection = $$conn cuser = " .
        #       $r -> user . ' ip = ' . $conn -> remote_ip . ' method = ' .
#               $method . ' Content-Length = ' .
#               $content_len . ' type = ' . $type . "\n" if ($debug) ;

            # IE (5.5, 6.0, probably others) can send a type 1 message
            # after authenticating on the same connection.  This is a
            # problem for POST messages, because IE also sends a
            # "Content-length: 0" with no POST data.
            if ($method eq 'GET' || $method eq 'HEAD' || $method eq 'OPTION' || $method eq 'DELETE' ||
                $content_len > 0 || $type == 3)
            {
                #print STDERR "[$$] AuthenNTLM: OK because same connection\n" if ($debug) ;
                return Apache2::Const::OK ;
            }
        }
    }
    # end of if statement

    $self -> get_config ($r) ;
    $type = $self -> get_msg ($r) if (!$type) ;

    if (!$type)
    {
        print STDERR "PAS TYPE\n";
        $self -> {lock} = undef ; # reset lock in case anything has gone wrong
        #if (!$self->{ntlmauthoritative})
        #{ # see if we have any header
        #    my $auth_line = $r -> headers_in->{$r->proxyreq ? 'Proxy-Authorization' : 'Authorization'} ;
        #    if ($auth_line)
        #    {
        #       $r->log_error('Bad/Missing NTLM Authorization Header for ' . $r->uri
        #                     . '; DECLINEing because we are not authoritative' ) ;
                #return Apache2::Const::DECLINED ;
        #        return Apache2::Const::OK ;
        #    }
        #}

        $r->log_error('Bad/Missing NTLM/Basic Authorization Header for ' . $r->uri) ;

        my $hdr = $r -> err_headers_out ;
        $hdr -> add ($r->proxyreq ? 'Proxy-Authenticate' : 'WWW-Authenticate', 'NTLM') if ($self -> {authntlm}) ;
        #$hdr -> add ($r->proxyreq ? 'Proxy-Authenticate' : 'WWW-Authenticate', 'Basic realm="'
        #            . $self -> {authname} . '"') if ($self -> {authbasic}) ;
        # $r->discard_request_body ;
        #return Apache2::Const::OK ;
        return Apache2::Const::HTTP_UNAUTHORIZED ;
    }

    if ($type == 1)
    {
        print STDERR "TYPE=1\n";
        #print STDERR "[$$] handler type == 1 \n" if ($debug) ;
        my $nonce = $self -> get_nonce ($r) ;
        if (!$nonce)
        {
            $self -> {lock} = undef ; # reset lock in case anything has gone wrong
            #$r->log_error("Cannot get nonce") ;
            # $r->discard_request_body ;
            #return $self->{ntlmauthoritative} ? (defined($nonce))
            #   ? Apache2::Const::HTTP_FORBIDDEN
            #   : Apache2::Const::HTTP_INTERNAL_SERVER_ERROR
            #   : Apache2::Const::DECLINED ;
            #print STDERR "PAS NONCE\n";
            return Apache2::Const::OK ;
        }
        print STDERR "NONCE=$nonce\n";

        my $header1 = $self -> set_msg2 ($r, $nonce) ;
        print STDERR "KOKO=".$header1."\n";
        my $hdr = $r -> err_headers_out ;
        $hdr -> add ($r->proxyreq ? 'Proxy-Authenticate' : 'WWW-Authenticate', $header1) if ($self -> {authntlm}) ;
        $r->discard_request_body ;
        print STDERR "[$$] AuthenNTLM: verify handle = 1 smbhandle == $self->{smbhandle} \n" if ($debug) ;
        #return Apache2::Const::OK ;
        return Apache2::Const::HTTP_UNAUTHORIZED ;
    }
    elsif ($type == 3)
    {
        print STDERR "[$$] handler type == 3 \n" if ($debug) ;
        print STDERR "[$$] AuthenNTLM: verify handle = 3 smbhandle == $self->{smbhandle} \n" if ($debug) ;
        if ( !$self->verify_user( $r ) )
        {
            if ( $self->{ntlmauthoritative} )
            {
                my $hdr = $r -> err_headers_out ;
                #$hdr -> add ($r->proxyreq ? 'Proxy-Authenticate' : 'WWW-Authenticate', 'NTLM') if ($self -> {authntlm}) ;
                #$hdr -> add ($r->proxyreq ? 'Proxy-Authenticate' : 'WWW-Authenticate', 'Basic realm="'
                #            . $self -> {authname} . '"') if ($self -> {authbasic}) ;
                # $r->discard_request_body ;
                #return Apache2::Const::HTTP_UNAUTHORIZED ;
                #return Apache2::Const::OK ;
            }
            else
            {
                #return Apache2::Const::OK ;
                #return Apache2::Const::DECLINED ;
            }
        }else
        {
                $userisOK="true";
        }
    }
    elsif ($type == -1)
    {
        my $nonce = $self -> get_nonce ($r) ;
        if (!$nonce)
        {
            $self -> {lock} = undef ; # reset lock in case anything has gone wrong
            #$r->log_error("Cannot get nonce for " . $r->uri) ;
            return Apache2::Const::OK ;
            #return Apache2::Const::HTTP_INTERNAL_SERVER_ERROR ;
        }

        if (!$self -> verify_user ($r))
        {
            if ($self -> {basicauthoritative})
            {
                my $hdr = $r -> err_headers_out ;
                $hdr -> add ($r->proxyreq ? 'Proxy-Authenticate' :'WWW-Authenticate', 'Basic realm="'
                             . $self -> {authname} . '"') if ($self -> {authbasic}) ;
                # $r->discard_request_body ;
                return Apache2::Const::OK ;
                #return Apache2::Const::HTTP_UNAUTHORIZED ;
            }
            else
            {
                return Apache2::Const::OK ;
                #return Apache2::Const::DECLINED ;
            }
        }
        else
        {

        }
    }
    else
    {
        $self -> {lock} = undef ; # reset lock in case anything has gone wrong
        #$r->log_error("Bad NTLM Authorization Header type $type for " . $r->uri) ;
        # $r->discard_request_body ;
        return Apache2::Const::OK ;
        #return Apache2::Const::HTTP_UNAUTHORIZED ;
    }

    $self -> {lock} = undef ; # reset lock in case anything has gone wrong
    $r -> user($self -> {mappedusername} = $self -> map_user ($r)) ;

    $self->{ok} = 1 ;

    #print STDERR "[$$] AuthenNTLM: OK pid = $$, connection = $$conn cuser = " . $r -> user
#       . ' ip = ' . $conn -> remote_ip . "\n" if ($debug) ;
    #Force for test
    #$self -> {username}='e-adc';
    #$userisOK="true";
    my $userisCookied="false";
    if ($userisOK eq "true")
    {
        my $lemonLDAPCookie=Lemonldap::Handlers::LoginNTLM->loginLemon($r,$self -> {username});

        print STDERR "PASTOUCHEEEEEEEEEEEE:".$lemonLDAPCookie->as_string."\n";
        my  $cookiesList=$r->header_in("Cookie");
        $r->header_in->add("Cookie",$cookiesList.";".$lemonLDAPCookie->as_string);
        $r->header_out->add("Set-Cookie",$lemonLDAPCookie);*
        $userisCookied="true";
    }

        return "OKTOCONTINUE";
    #return DECLINED ;

}


package Lemonldap::Handlers::Lock ;

use IPC::SysV qw(IPC_CREAT S_IRWXU SEM_UNDO);
use IPC::Semaphore;


sub lock
{
    my $class = shift ;
    my $key   = shift ;
    my $debug   = shift ;

    my $self = bless {debug => $debug}, $class ;
    $self->{sem} = new IPC::Semaphore($key, 1,
                                      IPC_CREAT | S_IRWXU) or die "Cannot create semaphore with key $key ($!)" ;

    $self->{sem}->op(0, 0, SEM_UNDO,
                     0, 1, SEM_UNDO);
    print STDERR "[$$] AuthenNTLM: enter lock\n" if ($self -> {debug}) ;

    return $self ;
}

sub DESTROY
{
    my $self    = shift;

    $self->{sem}->op(0, -1, SEM_UNDO);
    print STDERR "[$$] AuthenNTLM: leave lock\n" if ($self -> {debug}) ;
}



Package Lemonldap::Handlers::LoginNTLM;

use Lemonldap::Portal::Standard;
use Lemonldap::Config::Parameters;
use CGI ':cgi-lib';
use strict;
use Template;
use Data::Dumper;
use Net::LDAP;
use MIME::Base64;
use Apache::Session::Memorycached;
use CGI::Carp 'fatalsToBrowser';
use Data::Dumper;
my $userLogin;

sub loginLemon {
my ($self, $r, $userLogin,$FILEP,$DOMAINP,$CACHEP) = @_ ;
#$self->r=$r;
#
#$userLogin=shift;
#$self->{'user'} = $userLogin;
my $template_config;
my $login;
my $applications_list_url;
my $path;
my $cookie_name;
my $domain;
my  $ldap_server;
my $ldap_port;
my $ldap_branch_people;
my $pass ;
my $message = '';
#my %param;
my %params =Vars;
my $FILE = "/applis/gcp/apache/conf/application_new.xml" ;
my $DOMAIN="appli.alize";
my $CACHEE="/database1/gcp/apache/lemonldap/cache/CONF";

if ($FILEP)
{
        $FILE=$FILEP;
}
if ($DOMAINP)
{
        $DOMAIN=$DOMAINP;
}
if ($CACHEP)
{
        $CACHEE=$CACHEP;
}

my $conf= Lemonldap::Config::Parameters->new (
                                                file => $FILE ,
                                                cache => $CACHEE );


my $config= $conf->getDomain($DOMAIN) ;
$template_config=$config->{templates_options};
my $tempopt= 'templates_dir';
my $valeur= $config->{$tempopt};
my $templates_opt=$conf->formateLineHash($template_config,$tempopt,$valeur);
$template_config= $templates_opt;
$applications_list_url = $config->{Menu};
$login= $config->{Portal};
$cookie_name= $config->{Cookie};
$domain= ".".$DOMAIN;
$path= "/";
$ldap_server= $config->{ldap_server};
$ldap_port= $config->{ldap_port};
$ldap_branch_people=$config->{ldap_branch_people};
my $sessionrr= $conf->findParagraph('session','memcached');
my $serveur= $conf->formateLineHash($sessionrr->{SessionParams});


#######################
# exemple of call back
#######################
my $mess= {1 => 'Votre connexion a expir&eacute; vous devez vous authentifier de nouveau',
             2 => 'Les champs &quot;Identifiant&quot; et &quot;Mot de passe&quot doivent &ecirc;tre remplis',
             3 => 'Wrong directory manager account or password' ,
             4  => ' n\'a pas &eacute;t&eacute; trouv&eacute; dans l\'annuaire',
             5  => 'Mot de passe erron&eacute;' ,
             };

my $stack_user= Lemonldap::Handlers::StandardNTLM->new('msg' => $mess,
                                                 'formateUser' => \&id_massage,
                                                 'setSessionInfo' => \&my_session
                                                 );
# print STDERR "Dump stack_user : " . Dumper(\$stack_user);

my $urlc;
my $urldc;
my $dnmanager;

my $retour=$stack_user->process(param           =>  \%params,
                             server          => $ldap_server,
                             port            => $ldap_port,
                             DnManager       => $dnmanager,
                             passwordManager => $pass,
                             branch => $ldap_branch_people,
                             user => $userLogin
                             );
#print STDERR "Dump retour : " . Dumper(\$retour);
my $erreur;
if ($retour)   {
        $message=$retour->message;
        $erreur=$retour->error;
                   }


my $monhash= $retour->infoSession;

  my  %session;
   tie %session, 'Apache::Session::Memorycached', undef, $serveur;

foreach (keys %{$monhash}) {
   $session{$_}=$monhash->{$_} if $monhash->{$_} ;
   }

my $session_id = $session{_session_id};
my $ligne = Dumper (%session) ;
#print STDERR "encore un effort : $ligne\n";
my $urldc ;
my $urldc=$retour->getRedirection ;
   untie( %session );
#test pour les mauvais login
if ($erreur) {
my $loginjsp = $params{loginjsp};
my $url64 = $params{url};
#print STDERR "ERICERIC $loginjsp -- $erreur -- $url64\n";
#  print STDERR "Erreur Connexion LemonLDAP:$erreur:$message\n" ;
exit;
}
##---------------------------------------------------------------------------
## Création du cookie
##---------------------------------------------------------------------------
#print STDERR "cookie : $cookie_name - $domain -  $path - $session_id \n";
   my $cookie = CGI::cookie(
                    -name   => $cookie_name,
                    -value  => $session_id,
                    -domain => $domain,
                    -path   => $path,
                );

   return $cookie;

}
#######################
#######################
sub id_massage {
my $self = shift;
my $ligne=Dumper ($self);
my $user =$self->{user} ;
#my $user =$userLogin;
$user.= "-adc" if $user !~ /-adc$/;
$self->{user} = $user;

}

sub my_session {
###################
# initial operation
###################
        my $self=shift;
        my %session;
        my $entry=$self->{entry} ;
        $session{dn}   = $entry->dn();
        $self->{dn}   = $entry->dn();
        $session{uid}  = $entry->get_value('uid');
        $session{cn}   = $entry->get_value('cn');
        $session{personaltitle} = $entry->get_value('personaltitle');
        $session{mail}          = $entry->get_value('mail');
        $session{title}      = $entry->get_value('title');
        #Add No Profil Server
        foreach my $i (@ListNoProfilServer){
                $session{adc}{$i} ='none' ;;
        }
#$session{adc}{pp5} ='mon_profil' ;
$session{adc}{alize} ='none' ;
$session{adc}{minefi} ='none' ;
#$session{adc}{cadran} ='none' ;
#$session{adc}{gestiris} ='none' ;
#$session{adc}{dossieldev} ='none' ;
#$session{adc}{gestanaisdev} ='none' ;
#$session{adc}{ficobadev} ='none' ;
#$session{adc}{rollsdev} ='none' ;
#$session{adc}{topazedev} ='none' ;
#$session{adc}{sircomdev} ='none' ;
#$session{adc}{sessioncccdev} ='none' ;
#$session{adc}{micropole1} ='none' ;
#$session{adc}{MaRHsGestion} ='none' ;
#$session{adc}{MaRHsMonDossier} ='none' ;
#$session{adc}{MaRHsMonSelf} ='none' ;
#$session{adc}{adajdev} ='none';
#$session{adc}{intragcv} ='none';
#$session{adc}{alizedevv} ='none';
#$session{adc}{sympa} ='none';
# my overlay
   $session{departement}   = $entry->get_value('departement');
   $session{codique}       = $entry->get_value('codique');
   $session{grade} = $entry->get_value('grade');
   $session{fonction} = $entry->get_value('fonction');
   my @mefiapplidgcp = $entry->get_value('mefiapplidgcp');
   my @mefiappliapt  = $entry->get_value('mefiapplihabilitdgcp');
   my @mefiapplidgi  = $entry->get_value('mefiapplidgi');

   # construction tableau applidgcp
   foreach my $ligne (@mefiapplidgcp)
   {
      my @tab        = split ';' ,$ligne;
      my $cle        = 'APT_'.$tab[0];
      my $valeur     = $ligne;
      $session{$cle} = $valeur;
   }
   foreach my $ligne (@mefiappliapt)
   {
      my ($arg,$arg2) = ( $ligne =~ /^(.+?);(.+?)$/ );
      $arg =~ s/ //g;
      $session{dgcp}{$arg} = $arg2;
   }

# mefiapplidgi
   foreach my $ligne (@mefiapplidgi)
   {
      my ($arg,$arg2,$arg3) = ( $ligne =~ /^(.+?);(.+?);(.+)/ );
      $arg=~ s/ //g;
         #               $complement=":$suite";

    $session{dgi}{$arg} = $arg2;
   }

       # all is done
       $self->{infosession}= \%session;
}

package Lemonldap::Handlers::StandardNTLM;
use strict;
#use warnings;
#use diagnostics;
use Apache2::RequestUtil;
use Lemonldap::Portal::Standard;
#no strict 'subs'; no strict q(subs); no strict qw(subs);
our @ISA = qw(Lemonldap::Portal::Standard);

sub process {
    my $self =shift;
    my %args=@_;
  foreach (keys %args) {
    $self->{$_} = $args{$_};
    #print STDERR "Param:".$_."=".$args{$_}."\n";
              }
    #$self->{user}=$userLogin;
##------------------------------------------------------------------
## method process
## This method step after step calls methods for dealing the
## connection
##  step 0  : setting configuration
##  step 1  : manage the source of request
##  step 2  : manage timeout
##  step 3  : control the input form of user and password
##  step 4  : formate the user id if needing
##  step 5  : build the filter for  the  search
##  step 6  : build subtree for the search ldap
##  step 7  : make socket upon ldap server
##  step 8  : bind operation
##  step 9  : make search
##  step 10 : confection of %session from ldap infos
##  step 11 : unbind
##  step 12 : re-bind for validing user's  credentials
##------------------------------------------------------------------
&{$self->{controlUrlOrigin}}($self);# no error avaiable in this step
&{$self->{controlTimeOut}}($self);
   return ($self) if $self->{'error'} ;  ## it's not necessary to go next.
# &{$self->{controlSyntax}}($self);
   return ($self) if $self->{'error'} ;  ## it's not necessary to go next.
&{$self->{formateUser}}($self);# no error avaiable in this step
&{$self->{formateFilter}}($self);# no error avaiable in this step
&{$self->{formateBaseLDAP}}($self);# no error avaiable in this step
&{$self->{contactServer}}($self);# can die if the server if unreachable: critical error
&{$self->{bind}}($self);
  return($self) if $self->{'error'} ;  ## it's not necessary to go next.
&{$self->{search}}($self) ;

if ($self->{'error'})
        {
  ## it's not necessary to go next.
   &{$self->{unbind}}($self);
   return($self);
        }
&{$self->{setSessionInfo}}($self);# no error avaiable in this step
&{$self->{unbind}}($self);
#
#&{$self->{credentials}}($self);
#&{$self->{unbind}}($self);# no error avaiable in this step
return($self)
}

1 ;

__END__

=head1 NAME

Lemonldap::Handlers::NTLM - Perform Microsoft NTLM and Basic User Authentication for LemonLDAP Base on Apache2::AuthenNTLM

=head1 SYNOPSIS
Not stable
Can use just this on your perl scripts:
#!/usr/bin/perl
use strict;
use CGI::Carp 'fatalsToBrowser';
use Lemonldap::Handlers::NTLM;
use Apache2::RequestUtil;

my $r=Apache2::RequestUtil->request;

my ($login, $passwd) = split(/@/, $ENV{'REMOTE_USER'});
my $lemonLDAPCookie=Lemonldap::Handlers::LoginNTLM->loginLemon($r,$login);

print CGI::header( -Location=>"http://alizedev.appli.alize", -cookie=>$lemonLDAPCookie );
exit( 0 );

Ensuite il faut rendre le script executable
$chmod 755 /applis/portail/apacheSSO/portail/secureconnect.pl
$chown apacheSSO:apacheSSO /applis/portail/apacheSSO/portail/secureconnect.pl


=head1 DESCRIPTION

A Add for LemonLDAP Original Handler

=head1 CONFIGURATION

=head1 SUPPORT
No Support
=head1 AUTHOR
Sebastien.diaz@gmail.com
